/*
 *
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"

#define PVR_EEPROM_I2C_ADDR    0x50

/*
 * Read and analyse data in the eeprom.  Use tveeprom to figure out
 * the packet structure, since this is another Hauppauge device and
 * internally it has a family resemblence to ivtv-type devices
 */

/* We seem to only be interested in the back half of the EEPROM */
#define EEPROM_SIZE 128
#define EEPROM_OFFS 128

int pvr_eeprom_analyse(struct pvr *pvr)
{
    u8 *eeprom;
    u8 offs;
    unsigned pcnt,tcnt;
    int ret;
    struct i2c_msg msg[2];

    memset(&pvr->tvdata,0,sizeof(pvr->tvdata));

    eeprom = kmalloc(EEPROM_SIZE,GFP_KERNEL);
    if (!eeprom) {
	return -ENOMEM;
    }

    msg[0].addr = PVR_EEPROM_I2C_ADDR;
    msg[0].flags = 0;
    msg[0].len = 1;
    msg[0].buf = &offs;
    msg[1].addr = PVR_EEPROM_I2C_ADDR;
    msg[1].flags = I2C_M_RD;

    /* We have to do the actual eeprom data fetch ourselves, because
       (1) we're only fetching part of the eeprom, and (2) if we were
       getting the whole thing our I2C driver can't grab it in one
       pass - which is what tveeprom is otherwise going to attempt */
    memset(eeprom,0,EEPROM_SIZE);
    for (tcnt = 0; tcnt < EEPROM_SIZE; tcnt += pcnt) {
	pcnt = 16;
	if (pcnt + tcnt > EEPROM_SIZE) pcnt = EEPROM_SIZE-tcnt;
	offs = tcnt + EEPROM_OFFS;
	msg[1].len = pcnt;
	msg[1].buf = eeprom+tcnt;
	if ((ret = i2c_transfer(&pvr->i2c_adap,
				msg,sizeof(msg)/sizeof(msg[0]))) != 2) {
	    pvr_debug("%s: eeprom fetch set offs err=%d",pvr->name,ret);
	    if (ret > 0) ret = -EIO;
	    kfree(eeprom);
	    return ret;
	}
    }

    tveeprom_hauppauge_analog(&pvr->tvdata,eeprom);

    pvr_debug("%s: eeprom results:",pvr->name);
    pvr_debug("%s: has_radio=%d",pvr->name,pvr->tvdata.has_radio);
    pvr_debug("%s: tuner_type=%d",pvr->name,pvr->tvdata.tuner_type);
    pvr_debug("%s: tuner_formats=0x%x",pvr->name,pvr->tvdata.tuner_formats);
    pvr_debug("%s: audio_processor=%d",pvr->name,pvr->tvdata.audio_processor);
    pvr_debug("%s: model=%d",pvr->name,pvr->tvdata.model);
    pvr_debug("%s: revision=%d",pvr->name,pvr->tvdata.revision);
    pvr_debug("%s: serial_number=%d",pvr->name,pvr->tvdata.serial_number);
    pvr_debug("%s: rev_str=%s",pvr->name,pvr->tvdata.rev_str);

    kfree(eeprom);

    return 0;
}
