/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-msp34x5g.h"
#include "msp3400.h"
#include <media/audiochip.h>

/*

  MCI <isely@pobox.com> 10-Mar-2005 - Rather than operate the msp34xx
  directly, we rely on the msp3400.ko module to do it for us.  We
  really have to do this because that $##@!! module is going to attach
  itself to us anyway, so we really can't operate the chip ourselves.
  Unfortunately msp3400.ko is a real train wreck of a piece of code.
  Most of the code below tries to tickle that module in just the right
  way to get the results we need.  Yuck.  msp3400.c should be taken
  out back and shot.  Based on my reading of the datasheet it should
  be possible to write a far cleaner and simpler driver than what is
  currently there right now.

*/

/* This function selects the correct audio input source */
int pvr_audio_set_stereo(struct pvr* pvr)
{
    int stat;
    unsigned short sarg = 0;
    struct msp_matrix mspm;

    if (pvr->audio.input == PVR_INPUT_TV) {
	struct v4l2_tuner vt;
	pvr_debug("pvr_audio_set_stereo(TV)");
	memset(&vt,0,sizeof(vt));
	vt.audmode = pvr->audio.mode;
	stat = pvr_i2c_msp3400_cmd(pvr,VIDIOC_S_TUNER,&vt);
	if (stat < 0) return stat;
    }

    sarg = AUDIO_TUNER;
    switch (pvr->audio.input) {
    case PVR_INPUT_TV:
	sarg = AUDIO_TUNER;
	break;
    case PVR_INPUT_RADIO:
	/* Assume that msp34xx also handle FM decoding, in which case
	   we're still using the tuner. */
	sarg = AUDIO_TUNER;
	break;
    case PVR_INPUT_COMPOSITE:
    case PVR_INPUT_SVIDEO:
	sarg = AUDIO_EXTERN;
	break;
    }
    stat = pvr_i2c_msp3400_cmd(pvr,AUDC_SET_INPUT,&sarg);
    if (stat < 0) return stat;

    /* The above should have been enough to do the job, however
       msp3400.ko does an incomplete job of handling the scart
       routing.  Really.  It doesn't even bother to initialize the SC1
       output at all.  So we have to help it here...  Unfortunately
       this also means that we have include msp3400.c's header file
       and it currently isn't in a public place.  Damnit! */

    mspm.input = SCART_IN1_DA;
    switch (pvr->audio.input) {
    case PVR_INPUT_COMPOSITE:
    case PVR_INPUT_SVIDEO:
	/* Bypass the DSP and just use IN1.  In theory we should be
	   able to permanent just use IN1_DA, but to do that
	   msp3400.ko should be adjusting the DSP input SCART routing
	   correctly when doing video in.  Unfortunately that appears
	   not to be the case.  I really hate that module. */
	mspm.input = SCART_IN1;
	break;
    }
    mspm.output = 1;
    stat = pvr_i2c_msp3400_cmd(pvr,MSP_SET_MATRIX,&mspm);
    if (stat < 0) return stat;

    return 0;
}

/* This sets the audio volume parameters */
int pvr_audio_setvolume(struct pvr *pvr)
{
    struct video_audio vt;
    pvr_debug("pvr_audio_setvolume");
    memset(&vt,0,sizeof(vt));
    vt.flags = (VIDEO_AUDIO_VOLUME |
		VIDEO_AUDIO_BALANCE |
		VIDEO_AUDIO_BASS |
		VIDEO_AUDIO_TREBLE |
		VIDEO_AUDIO_MUTABLE);
    if (pvr->audio.muted) vt.flags |= VIDEO_AUDIO_MUTE;
    vt.volume = pvr->audio.volume;
    vt.balance = pvr->audio.balance;
    vt.bass = pvr->audio.bass;
    vt.treble = pvr->audio.treble;
#ifdef notdef
    pvr_debug("%s: vol=%d bal=%d bas=%d treb=%d mute=%d",
	      pvr->name,pvr->audio.volume,pvr->audio.balance,
	      pvr->audio.bass,pvr->audio.treble,pvr->audio.muted);
#endif
    return pvr_i2c_msp3400_cmd(pvr,VIDIOCSAUDIO,&vt);
}

/* This reads back the current volume parameters and signal type */
int pvr_audio_update_status(struct pvr* pvr)
{
    struct video_audio vt;
    int stat;

    pvr_debug("pvr_audio_update_status");

    stat = pvr_i2c_msp3400_cmd(pvr,VIDIOCGAUDIO,&vt);
    if (stat < 0) return stat;

    if (vt.flags & VIDEO_AUDIO_MUTABLE) {
	pvr->audio.muted = (vt.flags & VIDEO_AUDIO_MUTE) != 0;
    }
    if (vt.flags & VIDEO_AUDIO_VOLUME) pvr->audio.volume = vt.volume;
    if (vt.flags & VIDEO_AUDIO_BASS) pvr->audio.bass = vt.bass;
    if (vt.flags & VIDEO_AUDIO_TREBLE) pvr->audio.treble = vt.treble;
    if (vt.flags & (VIDEO_AUDIO_BALANCE|VIDEO_AUDIO_VOLUME)) {
	pvr->audio.balance = vt.balance;
    }
    pvr->audio.status_stereo = (vt.mode & VIDEO_SOUND_STEREO) != 0;
    pvr->audio.status_bilingual = (vt.mode &
				   (VIDEO_SOUND_LANG1|VIDEO_SOUND_LANG2)) != 0;
    return 0;
}

/* This reconfigures everything about the sound processor (by calling
   the other functions.) */
int pvr_configure_audio(struct pvr *pvr)
{
    int ret;

    pvr_debug("pvr_configure_audio");

    /* Set stereo, */
    ret = pvr_audio_set_stereo(pvr);
    if (ret < 0) return ret;

    /* and finally... */
    ret = pvr_audio_setvolume(pvr);

    return ret;
}
