/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"

#define PVR_SAA7115_I2C_ADDR    0x21

/* These are sequences of register, value for the saa7115 chip,
 * Adapted from data of the ivtv project (saa7175.c, GPL2, Copyright owned by
 * 1998 Dave Perks <dperks@ibm.net>,
 * 2003 Ronald Bultje <rbultje@ronald.bitfreak.net>,
 * 2003 Kevin Thayer <nufan_wfk at yahoo.com>
 * 2004 Hans Verkuil <hverkuil@xs4all.nl>)
 * Comment is what hauppauge uses.
 */

static u8 decoder_set_auto_input[] = {
    0x01, 0x08, // 0x48
    0x03, 0x20,
    0x06, 0xEB,
    0x07, 0xE0,

    0x0a, 0x80, // 0x6e
    0x0b, 0x44, // 0x40
    0x0c, 0x40,
    0x0d, 0x00,
    0x0f, 0x00, // 0x80

    0x10, 0x06, // 0x00
    0x11, 0x00,
    0x12, 0x9d,
    0x13, 0x80,
    0x14, 0x00,
    0x15, 0x04, // 0x01, 0x36
    0x17, 0x98, // 0x99
    0x18, 0x40,
    0x19, 0x80,
    0x1a, 0x77,
    0x1b, 0x42,
    0x1c, 0xa9,
    0x1d, 0x01,
    0x88, 0xd0,
    0x88, 0xf0, 

    0x00, 0x00
};

static u8 decoder_reset_scaler[] = {
    0x87, 0x00,
    0x88, 0xd0,
    0x88, 0xf0,
    0x87, 0x01,

    0x00, 0x00
};

static u8 decoder_set_video_pal[] = {
0x80, 0x00, // reset tasks
0x88, 0xD0, // reset scaler

0x15, 0x37, // VGATE start
0x16, 0x16, // VGATE stop
0x17, 0x99, // VGATE MSB and other values

0x08, 0x28, // 0x28 = PAL
0x0E, 0x07, // chrominance control 1

0x5A, 0x03, // Vertical offset, standard PAL value

// Task A
0x90, 0x81, // Task Handling Control
0x91, 0x48, // X-port formats/config
0x92, 0x40, // Input Ref. signal Def.
0x93, 0x84, // I-port config
// This is weird: the datasheet says that you should use 2 as the minimum value,
// but Hauppauge uses 0, and changing that to 2 causes indeed problems (for PAL)
0x94, 0x00, // hoffset low (input), 0x0002 is minimum
0x95, 0x00, // hoffset hi (input)
0x96, 0xD0, // hsize low (input), 0x02d0 = 720
0x97, 0x02, // hsize hi (input)
0x98, 0x03, // voffset low (input)
0x99, 0x00, // voffset hi (input)
0x9A, 0x12, // vsize low (input), 0x12 = 18
0x9B, 0x00, // vsize hi (input)
0x9C, 0xA0, // hsize low (output), 0x05a0 = 1440
0x9D, 0x05, // hsize hi (output)
0x9E, 0x12, // vsize low (output), 0x12 = 18
0x9F, 0x00, // vsize hi (output)

// Task B
0xC0, 0x00, // Task Handling Control
0xC1, 0x08, // X-port formats/config
0xC2, 0x00, // Input Ref. signal Def.
0xC3, 0x80, // I-port config
0xC4, 0x00, // hoffset low (input), 0x0002 is minimum. See comment at 0x94 above.
0xC5, 0x00, // hoffset hi (input)
0xC6, 0xD0, // hsize low (input), 0x02d0 = 720
0xC7, 0x02, // hsize hi (input)
0xC8, 0x16, // voffset low (input), 0x16 = 22
0xC9, 0x00, // voffset hi (input)
0xCA, 0x20, // vsize low (input), 0x0120 = 288
0xCB, 0x01, // vsize hi (input)                       
0xCC, 0xD0, // hsize low (output), 0x02d0 = 720
0xCD, 0x02, // hsize hi (output)
0xCE, 0x20, // vsize low (output), 0x0120 = 288
0xCF, 0x01, // vsize hi (output)

0xF0, 0xB0, // Set PLL Register. PAL 625 lines per frame, 27 MHz
0xF1, 0x05, // low bit with 0xF0, (was 0x05)
0xF5, 0xB0, // Set pulse generator register
0xF6, 0x01,

0x87, 0x00, // Disable I-port output
0x88, 0xD0, // reset scaler (was 0xD0)
0x80, 0x20, // Activate only task "B"
0x88, 0xF0, // activate scaler
0x87, 0x01, // Enable I-port output
0x00, 0x00
};

static u8 decoder_set_video_ntsc[] = {
0x80, 0x00, // reset tasks
0x88, 0xD0, // reset scaler

0x15, 0x03, // VGATE pulse start
0x16, 0x11, // VGATE pulse stop
0x17, 0x9C, // VGATE MSB and other values

0x08, 0x68, // 0xBO: auto detection, 0x68 = NTSC
0x0E, 0x09, // lots of different stuff... video autodetection is on

0x5A, 0x06, // Vertical offset, standard NTSC value for ITU656 line counting

// Task A
0x90, 0x80, // Task Handling Control
0x91, 0x48, // X-port formats/config
0x92, 0x40, // Input Ref. signal Def.
0x93, 0x84, // I-port config
0x94, 0x01, // hoffset low (input), 0x0002 is minimum
0x95, 0x00, // hoffset hi (input)
0x96, 0xD0, // hsize low (input), 0x02d0 = 720
0x97, 0x02, // hsize hi (input)
0x98, 0x05, // voffset low (input)
0x99, 0x00, // voffset hi (input)
0x9A, 0x0C, // vsize low (input), 0x0c = 12
0x9B, 0x00, // vsize hi (input)
0x9C, 0xA0, // hsize low (output), 0x05a0 = 1440
0x9D, 0x05, // hsize hi (output)
0x9E, 0x0C, // vsize low (output), 0x0c = 12
0x9F, 0x00, // vsize hi (output)

// Task B
0xC0, 0x00, // Task Handling Control
0xC1, 0x08, // X-port formats/config
0xC2, 0x00, // Input Ref. signal Def.
0xC3, 0x80, // I-port config
0xC4, 0x02, // hoffset low (input), 0x0002 is minimum
0xC5, 0x00, // hoffset hi (input)
0xC6, 0xD0, // hsize low (input), 0x02d0 = 720
0xC7, 0x02, // hsize hi (input)
0xC8, 0x12, // voffset low (input), 0x12 = 18
0xC9, 0x00, // voffset hi (input)
0xCA, 0xF8, // vsize low (input), 0xf8 = 248
0xCB, 0x00, // vsize hi (input)
0xCC, 0xD0, // hsize low (output), 0x02d0 = 720
0xCD, 0x02, // hsize hi (output)

0xF0, 0xAD, // Set PLL Register. NTSC 525 lines per frame, 27 MHz
0xF1, 0x05, // low bit with 0xF0
0xF5, 0xAD, // Set pulse generator register
0xF6, 0x01,

0x87, 0x00, // Disable I-port output
0x88, 0xD0, // reset scaler
0x80, 0x20, //Activate only task "B", continuous mode (was 0xA0)
0x88, 0xF0, //activate scaler
0x87, 0x01, //Enable I-port output
0x00, 0x00
};

static u8 decoder_set_misc[] = {
0x38, 0x03, // audio stuff
0x39, 0x10,
0x3A, 0x00,

0x81, 0x01, // reg 0x15,0x16 define blanking window
0x82, 0x00,
0x83, 0x01, // I port settings
0x84, 0x20,
0x85, 0x21,
0x86, 0xC5,
0x87, 0x01,

// Task A
0xA0, 0x01, // down scale = 1
0xA1, 0x00, // prescale accumulation length = 1
0xA2, 0x00, // dc gain and fir prefilter control
0xA4, 0x80, // Lum Brightness, nominal value = 0x80
0xA5, 0x40, // Lum contrast, nominal value = 0x40
0xA6, 0x40, // Chroma satur. nominal value = 0x80
0xA8, 0x00, // hor lum scaling 0x0200 = 2 zoom
0xA9, 0x02, // note: 2 x zoom ensures that VBI lines have same length as video lines.
0xAA, 0x00, // H-phase offset Luma = 0
0xAC, 0x00, // hor chrom scaling 0x0200. must be hor lum scaling /2
0xAD, 0x01, // H-scaling incr chroma 
0xAE, 0x00, // H-phase offset chroma. must be offset luma /2         

0xB0, 0x00, // V-scaling incr luma low
0xB1, 0x04, // " hi
0xB2, 0x00, // V-scaling incr chroma low
0xB3, 0x04, // " hi
0xB4, 0x01, // V-scaling mode control
0xB8, 0x00, // V-phase offset chroma 00
0xB9, 0x00, // V-phase offset chroma 01
0xBA, 0x00, // V-phase offset chroma 10
0xBB, 0x00, // V-phase offset chroma 11
0xBC, 0x00, // V-phase offset luma 00
0xBD, 0x00, // V-phase offset luma 01
0xBE, 0x00, // V-phase offset luma 10
0xBF, 0x00, // V-phase offset luma 11

// Task B
0xD0, 0x01, // down scale = 1
0xD1, 0x00, // prescale accumulation length = 1
0xD2, 0x00, // dc gain and fir prefilter control
0xD4, 0x80, // Lum Brightness, nominal value = 0x80
0xD5, 0x40, // Lum contrast, nominal value = 0x40
0xD6, 0x40, // Chroma satur. nominal value = 0x80
0xD8, 0x00, // hor lum scaling 0x0400 = 1
0xD9, 0x04,
0xDA, 0x00, // H-phase offset Luma = 0
0xDC, 0x00, // hor chrom scaling 0x0200. must be hor lum scaling /2
0xDD, 0x02, // H-scaling incr chroma
0xDE, 0x00, // H-phase offset chroma. must be offset luma /2

0xE0, 0x00, // V-scaling incr luma low
0xE1, 0x04, // " hi
0xE2, 0x00, // V-scaling incr chroma low
0xE3, 0x04, // " hi
0xE4, 0x01, // V-scaling mode control
0xE8, 0x00, // V-phase offset chroma 00
0xE9, 0x00, // V-phase offset chroma 01
0xEA, 0x00, // V-phase offset chroma 10
0xEB, 0x00, // V-phase offset chroma 11
0xEC, 0x00, // V-phase offset luma 00
0xED, 0x00, // V-phase offset luma 01
0xEE, 0x00, // V-phase offset luma 10
0xEF, 0x00, // V-phase offset luma 11

0xF2, 0x50, // crystal clock = 24.576 MHz, target = 27MHz
0xF3, 0x46,
0xF4, 0x00,
0xF7, 0x4B, // not the recommended settings!
0xF8, 0x00,
0xF9, 0x4B,
0xFA, 0x00,
0xFB, 0x4B,
0xFF, 0x88, // PLL2 lock detection settings: 71 lines 50% phase error

/* Turn off VBI */
0x40, 0x00,
0x41, 0xFF,
0x42, 0xFF,
0x43, 0xFF,
0x44, 0xFF,
0x45, 0xFF,
0x46, 0xFF,
0x47, 0xFF,
0x48, 0xFF,
0x49, 0xFF,
0x4A, 0xFF,
0x4B, 0xFF,
0x4C, 0xFF,
0x4D, 0xFF,
0x4E, 0xFF,
0x4F, 0xFF,
0x50, 0xFF,
0x51, 0xFF,
0x52, 0xFF,
0x53, 0xFF,
0x54, 0xFF,
0x55, 0xFF,
0x56, 0xFF,
0x57, 0xFF,
0x58, 0x40,
0x59, 0x47,
0x5B, 0x83,
0x5D, 0xBD,
0x5E, 0x35,

0x02, 0x84, // input tuner -> input 4, amplifier active
0x09, 0x53, // was 0x56 for NTSC. luminance control

0x80, 0x20, // enable task B
0x88, 0xD0,
0x88, 0xF0,
0x00, 0x00 
};

static u8 decoder_set_audio_48_pal[] = {
    0x34, 0xce,
    0x35, 0xfb,
    0x36, 0x30,
    0x30, 0x00,
    0x31, 0xc0,
    0x32, 0x03,
    0x00, 0x00
};

static u8 decoder_set_audio_44_pal[] = {
    0x34, 0xf2,
    0x35, 0x00,
    0x36, 0x2D,
    0x30, 0x00,
    0x31, 0x72,
    0x32, 0x03,
    0x00, 0x00
};

static u8 decoder_set_audio_48_ntsc[] = {
    0x34, 0xce,
    0x35, 0xfb,
    0x36, 0x30,
    0x30, 0xcd,
    0x31, 0x20,
    0x32, 0x03,
    0x00, 0x00
};

static u8 decoder_set_audio_44_ntsc[] = {
    0x34, 0xf2,
    0x35, 0x00,
    0x36, 0x2D,
    0x30, 0xbc,
    0x31, 0xdf,
    0x32, 0x02,
    0x00, 0x00
};

static u8 decoder_vbi_on[] = {
    0x80, 0x00,
    0x88, 0xD0,
    0x80, 0x30,
    0x88, 0xF0,
    0x87, 0x01,
    0x00, 0x00
};

/*
static u8 decoder_vbi_off[] = {
    0x80, 0x00,
    0x88, 0xD0,
    0x80, 0x20,
    0x88, 0xF0,
    0x87, 0x01,
    0x00, 0x00
};
*/

static int pvr_write_decoder_cmd(struct pvr *pvr, u8 cmd, u8 data)
{
    int ret;
    struct i2c_msg msg[1];
    u8 buff[] = { cmd, data };

//    pvr_debug("pvr_write_decoder_cmd");
    msg[0].addr = PVR_SAA7115_I2C_ADDR;
    msg[0].flags = 0;
    msg[0].len = sizeof(buff);
    msg[0].buf = buff;

    ret = i2c_transfer(&pvr->i2c_adap,msg,sizeof(msg)/sizeof(msg[0]));
    if (ret < 0) return ret;
    return 0;
}

static int pvr_write_decoder_cmds(struct pvr *pvr, u8 *cmd)
{
    int ret;
    u8 c,d;

//    pvr_debug("pvr_write_decoder_cmds");

    while (*cmd != 0x00) {
	c = *cmd++;
	d = *cmd++;
	ret = pvr_write_decoder_cmd(pvr,c,d);
	if (ret < 0) return ret;
    }

    return 0;
}

static int pvr_read_decoder_cmd(struct pvr *pvr, u8 addr, u8 *res)
{
    struct i2c_msg msg[2];
    int ret;

//    pvr_debug("pvr_read_decoder_cmd");

    msg[0].addr = PVR_SAA7115_I2C_ADDR;
    msg[0].flags = 0;
    msg[0].len = sizeof(addr);
    msg[0].buf = &addr;
    msg[1].addr = PVR_SAA7115_I2C_ADDR;
    msg[1].flags = I2C_M_RD;
    msg[1].len = sizeof(*res);
    msg[1].buf = res;

    ret = i2c_transfer(&pvr->i2c_adap,msg,sizeof(msg)/sizeof(msg[0]));
    if (ret < 0) return ret;
    return 0;
}

int pvr_decoder_set_input(struct pvr *pvr)
{
    switch(pvr->video.input){
        case PVR_INPUT_TV:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x84);
        case PVR_INPUT_COMPOSITE:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x85);
        case PVR_INPUT_SVIDEO:
            return pvr_write_decoder_cmd(pvr, 0x02, 0x88);
    }
    return -EINVAL;
}

int pvr_decoder_set_bcsh(struct pvr *pvr)
{
    int ret = 0;

    ret |= pvr_write_decoder_cmd(pvr, 0x0a, pvr->video.brightness);
    ret |= pvr_write_decoder_cmd(pvr, 0x0b, pvr->video.contrast);
    ret |= pvr_write_decoder_cmd(pvr, 0x0c, pvr->video.saturation);
    ret |= pvr_write_decoder_cmd(pvr, 0x0d, (s8) pvr->video.hue);

    return ret;
}

int pvr_decoder_set_size(struct pvr *pvr)
{
    int ret = 0;
    int is_ntsc = (pvr->video.standard == PVR_STD_NTSC);
    int height, full_height = (is_ntsc ? 480 : 576);
    int width = pvr->video.width;
    int hpsc, hfsc, vscy;

    pvr_debug("pvr_decoder_set_size");

    ret |= pvr_write_decoder_cmd(pvr, 0xcc, width & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xcd, (width >> 8) & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xce, full_height & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xcf, (full_height >> 8) & 0xff);

    hpsc = (int) (720 / width);
    hfsc = (int) (1024 * 720 / (width * hpsc));
    ret |= pvr_write_decoder_cmd(pvr, 0xd0, hpsc & 0x3f);
    /* H fine scaling luminance */
    ret |= pvr_write_decoder_cmd(pvr, 0xd8, hfsc & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xd9, (hfsc >> 8) & 0xff);
    /* H fine scaling chrominance (must be lum/2) */
    ret |= pvr_write_decoder_cmd(pvr, 0xdc, (hfsc >> 1) & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xdd, (hfsc >> 9) & 0xff);
    
    /* Vertical scaling will be done by mpeg encoder */
    height = full_height;
    vscy = (int) ((1024 * full_height) / height);
    /* Crontrast an d Luminance*/
    ret |= pvr_write_decoder_cmd(pvr, 0xd5, (u8) (64 * 1024 / vscy));
    ret |= pvr_write_decoder_cmd(pvr, 0xd6, (u8) (64 * 1024 / vscy));
    /* V fine-scaling (luminance)*/
    ret |= pvr_write_decoder_cmd(pvr, 0xe0, vscy & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xe1, (vscy >> 8) & 0xff);
    /* V fine-scaling (chrominance)*/
    ret |= pvr_write_decoder_cmd(pvr, 0xe2, vscy & 0xff);
    ret |= pvr_write_decoder_cmd(pvr, 0xe3, (vscy >> 8) & 0xff);

    return ret;
}

int pvr_configure_decoder(struct pvr *pvr)
{
    int ret = 0;
    int is_ntsc = (pvr->video.standard == PVR_STD_NTSC);

    /* Set some default values then set the right resolution. */
    
    ret |= pvr_write_decoder_cmds(pvr, decoder_set_auto_input);
    ret |= pvr_write_decoder_cmds(pvr, decoder_set_misc);
    ret |= pvr_decoder_set_size(pvr);
    ret |= pvr_decoder_set_bcsh(pvr);
    
    if (is_ntsc){
        ret |= pvr_write_decoder_cmds(pvr, decoder_set_video_ntsc);
        switch(pvr->audio.srate){
            case PVR_SRATE_48:
                ret |= pvr_write_decoder_cmds(pvr, decoder_set_audio_48_ntsc);
                break;
            case PVR_SRATE_44:
                ret |= pvr_write_decoder_cmds(pvr, decoder_set_audio_44_ntsc);
                break;
        }
    } else {
        ret |= pvr_write_decoder_cmds(pvr, decoder_set_video_pal);
        switch(pvr->audio.srate){
            case PVR_SRATE_48:
                ret |= pvr_write_decoder_cmds(pvr, decoder_set_audio_48_pal);
                break;
            case PVR_SRATE_44:
                ret |= pvr_write_decoder_cmds(pvr, decoder_set_audio_44_pal);
                break;
        }
    }
    ret |= pvr_write_decoder_cmds(pvr, decoder_reset_scaler);

    if (pvr->config == _vbi) {
        int i;
        u8 lcr[22];

        for (i = 2; i <= 23; i++)
            lcr[i - 2] = 0x00;

        if (is_ntsc)
            for (i = 10; i <= 21; i++)
                lcr[i - 2] = 0xdd;
        else
             for (i = 6; i <= 23; i++)
                lcr[i - 2] = 0xdd;

        for (i = 2; i <= 23; i++)
            ret |= pvr_write_decoder_cmd(pvr, 0x41 + i - 2, lcr[i - 2]);

        ret |= pvr_write_decoder_cmds(pvr, decoder_vbi_on);
    }

    return ret;
}

int pvr_decoder_is_tuned(struct pvr *pvr)
{
    u8 status, ret =0;

    ret = pvr_read_decoder_cmd(pvr, 0x1f, &status);

    return (ret == 0) && !(status & (1 << 6));
}
